@@ -1065,6 +1065,7 @@ bit_write_BE (Bit_Chain *dat, double x, double y, double z)
     }
 }
 
+// TODO ensure length 1, dwg_geom_normalize()
 void
 normalize_BE (BITCODE_3BD ext)
 {
@@ -1366,6 +1367,8 @@ bit_write_CRC (Bit_Chain *dat, long unsigned int start_address, uint16_t seed)
   while (dat->bit > 0)
     bit_write_B (dat, 0);
 
+  if (start_address > dat->byte || (dat->byte + 2) >= dat->size)
+      bit_chain_alloc (dat);
   if (start_address > dat->byte || dat->byte >= dat->size)
     {
       loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
@@ -1391,6 +1394,8 @@ bit_write_CRC_LE (Bit_Chain *dat, long unsigned int start_address, uint16_t seed
   while (dat->bit > 0)
     bit_write_B (dat, 0);
 
+  if (start_address > dat->byte || (dat->byte + 2) >= dat->size)
+      bit_chain_alloc (dat);
   if (start_address > dat->byte || dat->byte >= dat->size)
     {
       loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
@@ -1441,14 +1446,15 @@ ATTRIBUTE_MALLOC
 BITCODE_TF
 bit_read_TF (Bit_Chain *restrict dat, unsigned int length)
 {
-  BITCODE_RC *chain = (BITCODE_RC *)calloc (length + 1, 1);
+  BITCODE_RC *chain;
+  CHK_OVERFLOW_PLUS (length,__FUNCTION__,NULL)
+  chain = (BITCODE_RC *)calloc (length + 1, 1);
   if (!chain)
     {
       loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
       LOG_ERROR ("Out of memory");
       return NULL;
     }
-
   bit_read_fixed (dat, chain, length);
   chain[length] = '\0';
 
@@ -1465,7 +1471,9 @@ bit_read_bits (Bit_Chain *dat, unsigned long bits)
 {
   unsigned bytes = bits / 8;
   int rest = bits % 8;
-  BITCODE_RC *restrict chain = (BITCODE_RC *)calloc (bytes + (rest ? 2 : 1), 1);
+  BITCODE_RC *restrict chain;
+  CHK_OVERFLOW_PLUS (bytes,__FUNCTION__,NULL)
+  chain = (BITCODE_RC *)calloc (bytes + (rest ? 2 : 1), 1);
   if (!chain)
     {
       loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
@@ -1542,8 +1550,9 @@ static int heX (unsigned char c)
   return c >= 10 ? 'A' + c - 10 : '0' + c;
 }
 
-/* converts TU to ASCII with embedded \U+XXXX.
-   max len widechars.
+/* Converts TU to ASCII with embedded \U+XXXX.
+   Max len widechars.
+   No codepage support yet, only cp 30.
  */
 char *
 bit_embed_TU_size (BITCODE_TU restrict wstr, const int len)
@@ -1813,7 +1822,7 @@ bit_write_T (Bit_Chain *restrict dat, BITCODE_T restrict s)
 {
   int i, length;
 
-  // only if from r2007+ DWG, not JSON, DXF
+  // only if from r2007+ DWG, not JSON, DXF, add API
   if (IS_FROM_TU (dat))
     {
       // downconvert TU to TV
@@ -1905,6 +1914,32 @@ bit_read_TU (Bit_Chain *restrict dat)
   return chain;
 }
 
+BITCODE_TU
+bit_read_TU_len (Bit_Chain *restrict dat, unsigned int *lenp)
+{
+  unsigned int i;
+  unsigned int length;
+  BITCODE_TU chain;
+
+  CHK_OVERFLOW_PLUS (1,__FUNCTION__,NULL)
+  length = bit_read_BS (dat);
+  CHK_OVERFLOW_PLUS (length * 2,__FUNCTION__,NULL)
+  chain = (BITCODE_TU)malloc ((length + 1) * 2);
+  if (!chain)
+    {
+      loglevel = dat->opts & DWG_OPTS_LOGLEVEL;
+      LOG_ERROR ("Out of memory")
+      return NULL;
+    }
+  for (i = 0; i < length; i++)
+    chain[i] = bit_read_RS (dat); // probably without byte swapping
+  // normally not needed, as the DWG itself contains the ending 0 as last char
+  // but we enforce writing it.
+  chain[length] = 0;
+  *lenp = length;
+  return chain;
+}
+
 /** String16: Read ASCII prefixed by a RS length
  */
 BITCODE_TV
@@ -2180,9 +2215,11 @@ bit_read_T (Bit_Chain *restrict dat)
     return (BITCODE_T)bit_read_TV (dat);
 }
 
-/* converts UCS-2 to UTF-8 */
+/* converts UCS-2LE to UTF-8.
+   first pass to get the dest len. single malloc.
+ */
 char *
-bit_convert_TU (BITCODE_TU restrict wstr)
+bit_convert_TU (const BITCODE_TU restrict wstr)
 {
   BITCODE_TU tmp = wstr;
   char *str;
@@ -2196,7 +2233,7 @@ bit_convert_TU (BITCODE_TU restrict wstr)
   if ((uintptr_t)wstr % SIZEOF_SIZE_T)
     {
       unsigned char *b = (unsigned char*)wstr;
-      c = (b[0] << 8) + b[1];
+      c = (b[1] << 8) + b[0];
       while (c)
         {
           len++;
@@ -2207,7 +2244,7 @@ bit_convert_TU (BITCODE_TU restrict wstr)
                 len++;
             }
           b += 2;
-          c = (b[0] << 8) + b[1];
+          c = (b[1] << 8) + b[0];
         }
     }
   else
@@ -2226,7 +2263,7 @@ bit_convert_TU (BITCODE_TU restrict wstr)
         LOG_INSANE ("U+%04X ", c);
 #endif
     }
-  str = (char*)malloc (len + 2);
+  str = (char*)malloc (len + 1);
   if (!str)
     {
       loglevel = 1;
@@ -2234,11 +2271,13 @@ bit_convert_TU (BITCODE_TU restrict wstr)
       return NULL;
     }
   i = 0;
+  tmp = wstr;
 #ifdef HAVE_ALIGNED_ACCESS_REQUIRED
   if ((uintptr_t)wstr % SIZEOF_SIZE_T)
     {
       unsigned char *b = (unsigned char*)wstr;
-      c = (b[0] << 8) + b[1];
+      // possible gcc bug
+      c = (b[1] << 8) + b[0];
       while (c && i < len)
         {
           if (c < 0x80)
@@ -2258,12 +2297,12 @@ bit_convert_TU (BITCODE_TU restrict wstr)
             }
 
           b += 2;
-          c = (b[0] << 8) + b[1];
+          c = (b[1] << 8) + b[0];
         }
     }
   else
 #endif
-  while ((c = *wstr++) && i < len)
+  while ((c = *tmp++) && i < len)
     {
       if (c < 0x80)
         {
@@ -2301,13 +2340,112 @@ bit_convert_TU (BITCODE_TU restrict wstr)
   return str;
 }
 
+#define EXTEND_SIZE(str, i, len)                \
+  if (i > len)                                  \
+    str = (char*) realloc (str, i+1)
+
+/* converts UCS-2LE to UTF-8. len is the wstr length, not the resulting utf8-size.
+   single pass with realloc. */
+char *
+bit_TU_to_utf8_len (const BITCODE_TU restrict wstr, const int len)
+{
+  BITCODE_TU tmp = wstr;
+  char *str;
+  int i = 0;
+  uint16_t c = 0;
+
+  if (!wstr || !len)
+    return NULL;
+  str = (char*)malloc (len + 1);
+  if (!str)
+    {
+      loglevel = 1;
+      LOG_ERROR ("Out of memory")
+      return NULL;
+    }
+  i = 0;
+  tmp = wstr;
+#ifdef HAVE_ALIGNED_ACCESS_REQUIRED
+  if ((uintptr_t)wstr % SIZEOF_SIZE_T)
+    {
+      unsigned char *b = (unsigned char*)wstr;
+      // possible gcc bug
+      c = (b[1] << 8) + b[0];
+      while (c && i < len)
+        {
+          if (c < 0x80)
+            {
+              str[i++] = c & 0xFF;
+            }
+          else if (c < 0x800)
+            {
+              EXTEND_SIZE(str, i + 1, len);
+              str[i++] = (c >> 6) | 0xC0;
+              str[i++] = (c & 0x3F) | 0x80;
+            }
+          else /* if (c < 0x10000) */
+            {
+              EXTEND_SIZE(str, i + 2, len);
+              str[i++] = (c >> 12) | 0xE0;
+              str[i++] = ((c >> 6) & 0x3F) | 0x80;
+              str[i++] = (c & 0x3F) | 0x80;
+            }
+
+          b += 2;
+          c = (b[1] << 8) + b[0];
+        }
+    }
+  else
+#endif
+  while ((c = *tmp++) && i < len)
+    {
+      if (c < 0x80)
+        {
+          str[i++] = c & 0xFF;
+        }
+      else if (c < 0x800)
+        {
+          EXTEND_SIZE(str, i + 1, len);
+          str[i++] = (c >> 6) | 0xC0;
+          str[i++] = (c & 0x3F) | 0x80;
+        }
+      else /* if (c < 0x10000) */
+        { /* windows ucs-2 has no D800-DC00 surrogate pairs. go straight up */
+          /*if (i+3 > len) {
+            str = realloc(str, i+3);
+            len = i+2;
+          }*/
+          EXTEND_SIZE(str, i + 2, len);
+          str[i++] = (c >> 12) | 0xE0;
+          str[i++] = ((c >> 6) & 0x3F) | 0x80;
+          str[i++] = (c & 0x3F) | 0x80;
+        }
+      /*
+      else if (c < 0x110000)
+        {
+          EXTEND_SIZE(str, i + 3, len);
+          str[i++] = (c >> 18) | 0xF0;
+          str[i++] = ((c >> 12) & 0x3F) | 0x80;
+          str[i++] = ((c >> 6) & 0x3F) | 0x80;
+          str[i++] = (c & 0x3F) | 0x80;
+        }
+      else
+        HANDLER (OUTPUT, "ERROR: overlarge unicode codepoint U+%0X", c);
+     */
+    }
+  if (i <= len + 1)
+    str[i] = '\0';
+  return str;
+}
+
 /** converts UTF-8 (dxf,json) to ASCII TV.
     Unquotes \" to ", \\ to \,
     undo json_cquote(), \uxxxx or other unicode => \U+XXXX.
     Returns NULL if not enough room in dest.
 */
 char *
-bit_utf8_to_TV (char *restrict dest, const unsigned char *restrict src, const int destlen)
+bit_utf8_to_TV (char *restrict dest, const unsigned char *restrict src,
+                const int destlen, const unsigned cquoted)
 {
   unsigned char c;
   unsigned char *s = (unsigned char *)src;
@@ -2320,7 +2458,7 @@ bit_utf8_to_TV (char *restrict dest, const unsigned char *restrict src, const in
         {
           return NULL;
         }
-      else if (c == '\\' && dest+1 < endp &&
+      else if (cquoted && c == '\\' && dest+1 < endp &&
           // skip \" to " and \\ to \.
           (*s == '"' || *s == '\\' || *s == 'r' || *s == 'n'))
         {
@@ -2411,7 +2549,7 @@ bit_utf8_to_TV (char *restrict dest, const unsigned char *restrict src, const in
     TODO: unquote json_cquote as above.
  */
 BITCODE_TU
-bit_utf8_to_TU (char *restrict str)
+bit_utf8_to_TU (char *restrict str, const unsigned cquoted)
 {
   BITCODE_TU wstr;
   int i = 0;
@@ -2471,7 +2609,7 @@ bit_utf8_to_TU (char *restrict str)
 int
 bit_eq_T (Bit_Chain *restrict dat, const BITCODE_T restrict wstr1, const char *restrict str2)
 {
-  if (dat->version >= R_2007)
+  if (IS_FROM_TU (dat))
     return bit_eq_TU (str2, (BITCODE_TU)wstr1);
   else
     return !strcmp (wstr1, str2);
@@ -2517,7 +2655,7 @@ bit_set_T (Bit_Chain *dat, const char* restrict src)
   if (!(IS_FROM_TU (dat)))
     return strdup (src);
   else
-    return (BITCODE_T)bit_utf8_to_TU ((char*)src);
+    return (BITCODE_T)bit_utf8_to_TU ((char*)src, 0);
 }
 
 /** Read 1 bitlong according to normal order
